"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.addCommand = exports.stackCommand = exports.searchCommand = exports.getTargetConfigAsync = exports.getAppTargetAsync = exports.listHardwareVariantsAsync = exports.initCommand = exports.installCommand = exports.buildCommandOnce = exports.resolveProject = exports.cleanCommand = exports.downloadCommand = exports.applyGlobalOptions = void 0;
const path = require("path");
const chalk = require("chalk");
const mkc = require("./mkc");
const files = require("./files");
const downloader = require("./downloader");
const loader_1 = require("./loader");
const mkc_1 = require("./mkc");
const stackresolver_1 = require("./stackresolver");
const files_1 = require("./files");
const host_1 = require("./host");
function downloadProjectAsync(id) {
    return __awaiter(this, void 0, void 0, function* () {
        id = id.replace(/.*\//, "");
        const url = mkc.cloudRoot + id + "/text";
        const files = yield downloader.httpGetJsonAsync(url);
        for (let fn of Object.keys(files)) {
            if (/\//.test(fn))
                continue;
            yield (0, host_1.host)().writeFileAsync(fn, files[fn]);
        }
        msg("downloaded.");
    });
}
function buildOnePrj(opts, prj) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const simpleOpts = {
                native: opts.native,
            };
            const res = yield prj.buildAsync(simpleOpts);
            const msgToString = (diagnostic) => {
                const category = diagnostic.category == 1
                    ? chalk.red("error")
                    : diagnostic.category == 2
                        ? chalk.yellowBright("warning")
                        : "message";
                return `${category} TS${diagnostic.code}: ${diagnostic.messageText}\n`;
            };
            let output = "";
            for (let diagnostic of res.diagnostics) {
                let pref = "";
                if (diagnostic.fileName)
                    pref = `${diagnostic.fileName}(${diagnostic.line + 1},${diagnostic.column + 1}): `;
                if (typeof diagnostic.messageText == "string")
                    output += pref + msgToString(diagnostic);
                else {
                    for (let chain = diagnostic.messageText; chain; chain = chain.next) {
                        output += pref + msgToString(chain);
                    }
                }
            }
            if (output)
                console.log(output.replace(/\n$/, ""));
            return res;
        }
        catch (e) {
            error("Exception: " + e.stack);
            return null;
        }
    });
}
function info(msg) {
    console.log(chalk.blueBright(msg));
}
function msg(msg) {
    console.log(chalk.green(msg));
}
function error(msg) {
    console.error(chalk.red(msg));
}
function applyGlobalOptions(opts) {
    if (opts.noColors)
        chalk.level = 0;
    else if (opts.colors && !chalk.level)
        chalk.level = 1;
    else if ((0, host_1.host)().getEnvironmentVariable("GITHUB_WORKFLOW"))
        chalk.level = 1;
}
exports.applyGlobalOptions = applyGlobalOptions;
function downloadCommand(URL, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        applyGlobalOptions(opts);
        yield downloadProjectAsync(URL);
    });
}
exports.downloadCommand = downloadCommand;
function cleanCommand(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        applyGlobalOptions(opts);
        for (const dir of ["built", "pxt_modules"]) {
            if (yield (0, host_1.host)().existsAsync(dir)) {
                msg(`deleting ${dir} folder`);
                yield (0, host_1.host)().rmdirAsync(dir, { recursive: true, force: true });
            }
        }
        msg("run `mkc init` again to setup your project");
    });
}
exports.cleanCommand = cleanCommand;
function resolveProject(opts, quiet = false) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const prjdir = yield files.findProjectDirAsync();
        if (!prjdir) {
            error(`could not find "pxt.json" file`);
            (0, host_1.host)().exitWithStatus(1);
        }
        if (!opts.configPath) {
            const cfgFolder = yield files.findParentDirWithAsync(prjdir, "mkc.json");
            if (cfgFolder)
                opts.configPath = path.join(cfgFolder, "mkc.json");
        }
        log(`using project: ${prjdir}/pxt.json`);
        const prj = new mkc.Project(prjdir);
        if (opts.configPath) {
            log(`using config: ${opts.configPath}`);
            prj.mkcConfig = yield readCfgAsync(opts.configPath, quiet);
        }
        yield prj.loadEditorAsync(!!opts.update);
        let version = "???";
        try {
            const appTarget = yield prj.service.languageService.getAppTargetAsync();
            version = (_a = appTarget === null || appTarget === void 0 ? void 0 : appTarget.versions) === null || _a === void 0 ? void 0 : _a.target;
        }
        catch (_b) { }
        log(`using editor: ${prj.mkcConfig.targetWebsite} v${version}`);
        if (opts.debug)
            yield prj.service.languageService.enableDebugAsync();
        if (opts.compileFlags) {
            yield prj.service.languageService.setCompileSwitchesAsync(opts.compileFlags);
        }
        prj.writePxtModules = !!opts.pxtModules;
        if (opts.linkPxtModules) {
            prj.writePxtModules = true;
            prj.linkPxtModules = true;
        }
        else if (opts.symlinkPxtModules) {
            prj.writePxtModules = true;
            prj.symlinkPxtModules = true;
        }
        return prj;
        function log(msg) {
            if (!quiet)
                info(msg);
        }
    });
}
exports.resolveProject = resolveProject;
function buildCommandOnce(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const prj = yield resolveProject(opts);
        yield prj.service.languageService.enableExperimentalHardwareAsync();
        const hwVariants = yield prj.service.getHardwareVariantsAsync();
        const appTarget = yield prj.service.languageService.getAppTargetAsync();
        const targetId = appTarget.id;
        let moreHw = [];
        const outputs = [];
        if (opts.hw) {
            const hws = opts.hw.split(/[\s,;]+/);
            selectHW(hws[0]);
            moreHw = hws.slice(1);
        }
        if (!opts.javaScript || opts.hw)
            opts.native = true;
        else
            opts.native = false;
        if (opts.native && hwVariants.length) {
            yield prj.guessHwVariantAsync();
            infoHW();
        }
        outputs.push(prj.outputPrefix);
        const compileRes = yield buildOnePrj(opts, prj);
        if (compileRes.success && opts.deploy) {
            const firmwareName = ["binary.uf2", "binary.hex", "binary.elf"].filter(f => !!compileRes.outfiles[f])[0];
            if (!firmwareName) {
                // something went wrong here
                error(`firmware missing from built files (${Object.keys(compileRes.outfiles).join(", ")})`);
            }
            else {
                const compileInfo = appTarget.compile;
                const drives = yield (0, host_1.host)().getDeployDrivesAsync(compileInfo);
                if (drives.length == 0) {
                    msg("cannot find any drives to deploy to");
                }
                else {
                    const firmware = compileRes.outfiles[firmwareName];
                    const encoding = firmwareName == "binary.hex" ? "utf8" : "base64";
                    msg(`copying ${firmwareName} to ` + drives.join(", "));
                    const writeHexFile = (drivename) => {
                        return (0, host_1.host)().writeFileAsync(path.join(drivename, firmwareName), firmware, encoding)
                            .then(() => info("   wrote to " + drivename))
                            .catch(() => error(`   failed writing to ${drivename}`));
                    };
                    for (const p of drives.map(writeHexFile))
                        yield p;
                }
            }
        }
        let success = compileRes.success;
        if (success && opts.monoRepo) {
            const dirs = yield (0, files_1.monoRepoConfigsAsync)(".");
            info(`mono-repo: building ${dirs.length} projects`);
            for (const fullpxtjson of dirs) {
                if (fullpxtjson.startsWith("pxt_modules"))
                    continue;
                const fulldir = path.dirname(fullpxtjson);
                info(`build ${fulldir}`);
                const prj0 = yield prj.mkChildProjectAsync(fulldir);
                const cfg = yield prj0.readPxtConfig();
                if (cfg.supportedTargets &&
                    cfg.supportedTargets.indexOf(targetId) < 0) {
                    info(`skipping due to supportedTargets`);
                    continue;
                }
                const res = yield buildOnePrj(opts, prj0);
                if (!res.success)
                    success = false;
            }
        }
        else if (success && moreHw.length) {
            for (const hw of moreHw) {
                selectHW(hw);
                infoHW();
                outputs.push(prj.outputPrefix);
                yield buildOnePrj(opts, prj);
            }
            const uf2s = [];
            for (const folder of outputs) {
                try {
                    uf2s.push(yield (0, host_1.host)().readFileAsync(path.join(folder, "binary.uf2")));
                }
                catch (_a) { }
            }
            if (uf2s.length > 1) {
                const total = concatUint8Arrays(uf2s);
                const fn = "built/combined.uf2";
                info(`combining ${uf2s.length} UF2 files into ${fn} (${Math.round(total.length / 1024)}kB)`);
                yield (0, host_1.host)().writeFileAsync(fn, total);
            }
        }
        if (success) {
            msg("Build OK");
            if (opts.watch)
                return compileRes;
            else
                (0, host_1.host)().exitWithStatus(0);
        }
        else {
            error("Build failed");
            if (opts.watch)
                return compileRes;
            else
                (0, host_1.host)().exitWithStatus(1);
        }
        return null;
        function hwid(cfg) {
            return cfg.name.replace(/hw---/, "");
        }
        function selectHW(hw0) {
            const hw = hw0.toLowerCase();
            const selected = hwVariants.filter(cfg => {
                return (cfg.name.toLowerCase() == hw ||
                    hwid(cfg).toLowerCase() == hw ||
                    cfg.card.name.toLowerCase() == hw);
            });
            if (!selected.length) {
                error(`No such HW id: ${hw0}`);
                msg(`Available hw:`);
                for (let cfg of hwVariants) {
                    msg(`${hwid(cfg)}, ${cfg.card.name} - ${cfg.card.description}`);
                }
                (0, host_1.host)().exitWithStatus(1);
            }
            prj.hwVariant = hwid(selected[0]);
        }
        function infoHW() {
            info(`using hwVariant: ${prj.mainPkg.mkcConfig.hwVariant} (target ${targetId})`);
            if (!opts.alwaysBuilt)
                prj.outputPrefix = "built/" + prj.mainPkg.mkcConfig.hwVariant;
        }
    });
}
exports.buildCommandOnce = buildCommandOnce;
function installCommand(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        applyGlobalOptions(opts);
        if (!(yield (0, host_1.host)().existsAsync("pxt.json"))) {
            error("missing pxt.json");
            (0, host_1.host)().exitWithStatus(1);
        }
        opts.pxtModules = true;
        const prj = yield resolveProject(opts);
        prj.mainPkg = null;
        if (opts.monoRepo) {
            const dirs = yield (0, files_1.monoRepoConfigsAsync)(".");
            info(`mono-repo: building ${dirs.length} projects`);
            for (const fullpxtjson of dirs) {
                if (fullpxtjson.startsWith("pxt_modules"))
                    continue;
                const fulldir = path.dirname(fullpxtjson);
                info(`install ${fulldir}`);
                const prj0 = yield prj.mkChildProjectAsync(fulldir);
                yield prj0.maybeWritePxtModulesAsync();
            }
        }
        else {
            yield prj.maybeWritePxtModulesAsync();
        }
    });
}
exports.installCommand = installCommand;
function initCommand(template, deps, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        applyGlobalOptions(opts);
        if (!(yield (0, host_1.host)().existsAsync("pxt.json"))) {
            if (opts.importUrl) {
                yield downloadProjectAsync(opts.importUrl);
            }
            else {
                if (!template) {
                    error("missing template");
                    (0, host_1.host)().exitWithStatus(1);
                }
                const target = loader_1.descriptors.find(t => t.id === template);
                if (!target) {
                    error(`template not found`);
                    (0, host_1.host)().exitWithStatus(1);
                }
                msg(`initializing project for ${target.name}`);
                msg("saving main.ts");
                yield (0, host_1.host)().writeFileAsync("main.ts", "// add code here", "utf8");
                msg("saving pxt.json");
                yield (0, host_1.host)().writeFileAsync("pxt.json", JSON.stringify({
                    name: "my-project",
                    version: "0.0.0",
                    files: ["main.ts"],
                    supportedTargets: [target.targetId],
                    dependencies: target.dependencies ||
                        (target.corepkg && { [target.corepkg]: "*" }) ||
                        {},
                    testDependencies: target.testDependencies || {},
                }, null, 4));
                yield (0, host_1.host)().writeFileAsync("mkc.json", JSON.stringify({
                    targetWebsite: target.website,
                    links: {},
                }, null, 4), "utf8");
            }
        }
        else {
            if (template || opts.importUrl) {
                error("directory is not empty, cannot apply template");
                (0, host_1.host)().exitWithStatus(1);
            }
        }
        const vscodeSettings = ".vscode/settings.json";
        if (opts.vscodeProject && !(yield (0, host_1.host)().existsAsync(vscodeSettings))) {
            if (!(yield (0, host_1.host)().existsAsync(".vscode")))
                yield (0, host_1.host)().mkdirAsync(".vscode");
            yield (0, host_1.host)().writeFileAsync(vscodeSettings, JSON.stringify({
                "editor.formatOnType": true,
                "files.autoSave": "afterDelay",
                "files.watcherExclude": {
                    "**/.git/objects/**": true,
                    "**/built/**": true,
                    "**/node_modules/**": true,
                    "**/yotta_modules/**": true,
                    "**/yotta_targets": true,
                    "**/pxt_modules/**": true,
                    "**/.pxt/**": true
                },
                "files.associations": {
                    "*.blocks": "html",
                    "*.jres": "json"
                },
                "search.exclude": {
                    "**/built": true,
                    "**/node_modules": true,
                    "**/yotta_modules": true,
                    "**/yotta_targets": true,
                    "**/pxt_modules": true,
                    "**/.pxt": true
                },
                "files.exclude": {
                    "**/pxt_modules": true,
                    "**/.pxt": true
                }
            }, null, 4));
        }
        const gitignore = ".gitignore";
        if (opts.gitIgnore && !(yield (0, host_1.host)().existsAsync(gitignore))) {
            msg(`saving ${gitignore}`);
            yield (0, host_1.host)().writeFileAsync(gitignore, `# MakeCode
built
node_modules
yotta_modules
yotta_targets
pxt_modules
.pxt
_site
*.db
*.tgz
.header.json
.simstate.json`);
        }
        if (!(yield (0, host_1.host)().existsAsync("tsconfig.json"))) {
            msg("saving tsconfig.json");
            yield (0, host_1.host)().writeFileAsync("tsconfig.json", JSON.stringify({
                compilerOptions: {
                    target: "ES5",
                    noImplicitAny: true,
                    outDir: "built",
                    rootDir: ".",
                },
                include: ["**/*.ts"],
                exclude: ["built/**", "pxt_modules/**/*test.ts"],
            }, null, 4), "utf8");
        }
        const prettierrc = ".prettierrc";
        if (!(yield (0, host_1.host)().existsAsync(prettierrc))) {
            msg(`saving ${prettierrc}`);
            yield (0, host_1.host)().writeFileAsync(prettierrc, JSON.stringify({
                arrowParens: "avoid",
                semi: false,
                tabWidth: 4,
            }));
        }
        const gh = ".github/workflows/makecode.yml";
        if (!(yield (0, host_1.host)().existsAsync(gh))) {
            if (!(yield (0, host_1.host)().existsAsync(".github")))
                yield (0, host_1.host)().mkdirAsync(".github");
            if (!(yield (0, host_1.host)().existsAsync(".github/workflows")))
                yield (0, host_1.host)().mkdirAsync(".github/workflows");
            msg(`saving ${gh}`);
            yield (0, host_1.host)().writeFileAsync(gh, `name: MakeCode Build
on:
  push:
  workflow_dispatch:

jobs:
  build:
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v3
        with:
          submodules: recursive
      - run: npx makecode
`);
        }
        opts.pxtModules = true;
        const prj = yield resolveProject(opts);
        if (!(yield (0, host_1.host)().existsAsync("mkc.json"))) {
            msg("saving mkc.json");
            yield (0, host_1.host)().writeFileAsync("mkc.json", mkc.stringifyConfig(prj.mainPkg.mkcConfig), "utf8");
        }
        for (const dep of deps)
            yield addDependency(prj, dep, undefined);
        prj.mainPkg = null;
        prj.writePxtModules = true;
        yield prj.maybeWritePxtModulesAsync();
        msg(`project ready, run "makecode -d" to build and deploy`);
    });
}
exports.initCommand = initCommand;
function listHardwareVariantsAsync(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const prj = yield resolveProject(opts);
        yield prj.service.languageService.enableExperimentalHardwareAsync();
        return yield prj.service.getHardwareVariantsAsync();
    });
}
exports.listHardwareVariantsAsync = listHardwareVariantsAsync;
function getAppTargetAsync(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const prj = yield resolveProject(opts);
        return yield prj.service.languageService.getAppTargetAsync();
    });
}
exports.getAppTargetAsync = getAppTargetAsync;
function getTargetConfigAsync(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const prj = yield resolveProject(opts);
        return yield prj.service.languageService.getTargetConfigAsync();
    });
}
exports.getTargetConfigAsync = getTargetConfigAsync;
function jacdacMakeCodeExtensions() {
    return __awaiter(this, void 0, void 0, function* () {
        let data = [];
        try {
            const r = yield (0, host_1.host)().requestAsync({
                url: "https://raw.githubusercontent.com/microsoft/jacdac/main/services/makecode-extensions.json"
            });
            data = JSON.parse(r.text);
        }
        catch (e) { }
        return data;
    });
}
function join(...parts) {
    return parts.filter(p => !!p).join("/");
}
// parse https://github.com/[company]/[project](/filepath)(#tag)
function parseRepoId(repo) {
    if (!repo)
        return undefined;
    // clean out whitespaces
    repo = repo.trim();
    // trim trailing /
    repo = repo.replace(/\/$/, "");
    // convert github pages into github repo
    const mgh = /^https:\/\/([^./#]+)\.github\.io\/([^/#]+)\/?$/i.exec(repo);
    if (mgh)
        repo = `github:${mgh[1]}/${mgh[2]}`;
    repo = repo.replace(/^github:/i, "");
    repo = repo.replace(/^https:\/\/github\.com\//i, "");
    repo = repo.replace(/\.git\b/i, "");
    const m = /^([^#\/:]+)\/([^#\/:]+)(\/([^#]+))?(#([^\/:]*))?$/.exec(repo);
    if (!m)
        return undefined;
    const owner = m[1];
    const project = m[2];
    let fileName = m[4];
    const tag = m[6];
    const treeM = fileName && /^tree\/([^\/]+\/)/.exec(fileName);
    if (treeM) {
        // https://github.com/pelikhan/mono-demo/tree/master/demo2
        fileName = fileName.slice(treeM[0].length);
        // branch info?
    }
    return {
        owner,
        project,
        slug: join(owner, project),
        fullName: join(owner, project, fileName),
        tag,
        fileName,
    };
}
function fetchExtension(slug) {
    return __awaiter(this, void 0, void 0, function* () {
        const url = `https://pxt.azureedge.net/api/gh/${slug}`;
        const req = yield (0, host_1.host)().requestAsync({
            url
        });
        if (req.statusCode !== 200) {
            error(`resolution of ${slug} failed (${req.statusCode})`);
            (0, host_1.host)().exitWithStatus(1);
        }
        const script = JSON.parse(req.text);
        return script;
    });
}
function searchCommand(query, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        applyGlobalOptions(opts);
        query = query.trim().toLowerCase();
        msg(`searching for ${query}`);
        const prj = yield resolveProject(opts);
        const targetid = prj.editor.targetJson.id;
        const res = yield (0, host_1.host)().requestAsync({
            url: `${mkc_1.cloudRoot}ghsearch/${targetid}/${targetid}?q=${encodeURIComponent(query)}`
        });
        if (res.statusCode !== 200) {
            error(`search request failed`);
            (0, host_1.host)().exitWithStatus(1);
        }
        const payload = JSON.parse(res.text);
        const { items } = payload;
        items === null || items === void 0 ? void 0 : items.forEach(({ full_name, description, owner }) => {
            msg(`  ${full_name}`);
            info(`    https://github.com/${full_name}`);
            if (description)
                info(`    ${description}`);
        });
        if (/jacdac/i.test(query)) {
            const q = query.replace(/jacdac-*/i, "");
            const exts = yield jacdacMakeCodeExtensions();
            for (const ext of exts.filter(ext => ext.client.name.indexOf(q) > -1 || ext.service.indexOf(q) > -1)) {
                msg(`  ${ext.client.name}`);
                info(`    https://${ext.client.repo}`);
            }
        }
    });
}
exports.searchCommand = searchCommand;
function stackCommand(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const srcmap = JSON.parse(yield (0, host_1.host)().readFileAsync("built/binary.srcmap", "utf8"));
        console.log((0, stackresolver_1.expandStackTrace)(srcmap, yield (0, host_1.host)().readFileAsync(0, "utf8")));
    });
}
exports.stackCommand = stackCommand;
function addCommand(pkg, name, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        applyGlobalOptions(opts);
        opts.pxtModules = true;
        msg(`adding ${pkg}`);
        const prj = yield resolveProject(opts);
        yield addDependency(prj, pkg, name);
        prj.mainPkg = null;
        yield prj.maybeWritePxtModulesAsync();
    });
}
exports.addCommand = addCommand;
function addDependency(prj, pkg, name) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        pkg = pkg.toLowerCase().trim();
        if (pkg === "jacdac")
            pkg = "https://github.com/microsoft/pxt-jacdac";
        else if (/^jacdac-/.test(pkg)) {
            const exts = yield jacdacMakeCodeExtensions();
            const ext = exts.find(ext => ext.client.name === pkg);
            if (ext) {
                info(`found jacdac ${ext.client.repo}`);
                pkg = ext.client.repo;
            }
        }
        const rid = parseRepoId(pkg);
        const pxtJson = yield prj.readPxtConfig();
        if (rid) {
            const d = yield fetchExtension(rid.slug);
            const dname = name ||
                join(rid.project, rid.fileName).replace(/^pxt-/, "").replace("/", "-");
            pxtJson.dependencies[dname] = `github:${rid.fullName}#${d.version ? `v${d.version}` : d.defaultBranch}`;
            info(`adding dependency ${dname}=${pxtJson.dependencies[dname]}`);
        }
        else {
            const appTarget = yield prj.service.languageService.getAppTargetAsync();
            const bundledPkgs = (_b = (_a = appTarget === null || appTarget === void 0 ? void 0 : appTarget.bundleddirs) === null || _a === void 0 ? void 0 : _a.map((dir) => { var _a; return (_a = /^libs\/(.+)/.exec(dir)) === null || _a === void 0 ? void 0 : _a[1]; })) === null || _b === void 0 ? void 0 : _b.filter((dir) => !!dir);
            const builtInPkg = bundledPkgs === null || bundledPkgs === void 0 ? void 0 : bundledPkgs.find(dir => dir === pkg);
            if (!builtInPkg) {
                const possiblyMeant = bundledPkgs === null || bundledPkgs === void 0 ? void 0 : bundledPkgs.filter(el => (el === null || el === void 0 ? void 0 : el.toLowerCase().indexOf(pkg)) !== -1);
                if (possiblyMeant === null || possiblyMeant === void 0 ? void 0 : possiblyMeant.length) {
                    error(`Did you mean ${possiblyMeant === null || possiblyMeant === void 0 ? void 0 : possiblyMeant.join(", ")}?`);
                }
                else {
                    error("unknown package, try https://github.com/.../... for github extensions");
                }
                (0, host_1.host)().exitWithStatus(1);
            }
            const collidingHwVariant = Object.keys(pxtJson.dependencies)
                .find(dep => dep.toLowerCase().replace(/---.+$/, "") === pkg.replace(/---.+$/, "")
                && pxtJson.dependencies[dep] === "*");
            if (collidingHwVariant) {
                delete pxtJson.dependencies[collidingHwVariant];
            }
            pxtJson.dependencies[builtInPkg] = "*";
            info(`adding builtin dependency ${builtInPkg}=*`);
        }
        yield (0, host_1.host)().writeFileAsync("pxt.json", JSON.stringify(pxtJson, null, 4), "utf8");
    });
}
function isKV(v) {
    return !!v && typeof v === "object" && !Array.isArray(v);
}
function jsonMergeFrom(trg, src) {
    if (!src)
        return;
    Object.keys(src).forEach(k => {
        if (isKV(trg[k]) && isKV(src[k]))
            jsonMergeFrom(trg[k], src[k]);
        else if (Array.isArray(trg[k]) && Array.isArray(src[k]))
            trg[k] = trg[k].concat(src[k]);
        else
            trg[k] = src[k];
    });
}
function readCfgAsync(cfgpath, quiet = false) {
    return __awaiter(this, void 0, void 0, function* () {
        const files = [];
        return readCfgRec(cfgpath);
        function readCfgRec(cfgpath) {
            return __awaiter(this, void 0, void 0, function* () {
                if (files.indexOf(cfgpath) >= 0) {
                    error(`Config file loop: ${files.join(" -> ")} -> ${cfgpath}`);
                    (0, host_1.host)().exitWithStatus(1);
                }
                const cfg = yield cfgFile(cfgpath);
                const currCfg = {};
                files.push(cfgpath);
                for (const fn of cfg.include || []) {
                    const resolved = path.resolve(path.dirname(cfgpath), fn);
                    if (!quiet)
                        info(`  include: ${resolved}`);
                    jsonMergeFrom(currCfg, readCfgRec(resolved));
                }
                jsonMergeFrom(currCfg, cfg);
                delete currCfg.include;
                files.pop();
                return currCfg;
                function cfgFile(cfgpath) {
                    return __awaiter(this, void 0, void 0, function* () {
                        let cfg;
                        try {
                            cfg = JSON.parse(yield (0, host_1.host)().readFileAsync(cfgpath, "utf8"));
                        }
                        catch (e) {
                            error(`Can't read config file: '${cfgpath}'; ` + e.message);
                            (0, host_1.host)().exitWithStatus(1);
                        }
                        const lnk = cfg.links;
                        if (lnk) {
                            const mkcFolder = path.resolve(".", path.dirname(cfgpath));
                            for (const k of Object.keys(lnk)) {
                                lnk[k] = path.resolve(mkcFolder, lnk[k]);
                            }
                        }
                        return cfg;
                    });
                }
            });
        }
    });
}
function concatUint8Arrays(bufs) {
    let size = 0;
    for (const buf of bufs) {
        size += buf.length;
    }
    const res = new Uint8Array(size);
    let offset = 0;
    for (const buf of bufs) {
        res.set(buf, offset);
        offset += buf.length;
    }
    return res;
}
//# sourceMappingURL=commands.js.map